package com.appian.ps.plugins.sftpsmartservices;

import java.nio.file.Paths;
import java.nio.file.Path;
import java.text.MessageFormat;
import java.util.Iterator;
import java.util.Locale;
import java.util.ResourceBundle;

import org.apache.log4j.Logger;

import com.appiancorp.suiteapi.common.Name;
import com.appiancorp.suiteapi.content.ContentService;
import com.appiancorp.suiteapi.knowledge.DocumentDataType;
import com.appiancorp.suiteapi.process.exceptions.SmartServiceException;
import com.appiancorp.suiteapi.process.framework.AppianSmartService;
import com.appiancorp.suiteapi.process.framework.Input;
import com.appiancorp.suiteapi.process.framework.Order;
import com.appiancorp.suiteapi.process.framework.Required;
import com.appiancorp.suiteapi.process.framework.SmartServiceContext;
import com.appiancorp.suiteapi.process.palette.ConnectivityServices;
import com.appiancorp.suiteapi.security.external.SecureCredentialsStore;

@ConnectivityServices
@Order({ "scsExternalSystemKey", "usePerUserCredentials", "privateKey",
		"knownHosts", "serverHostName", "serverPort", "remoteFilePath",
		"newDirectoryName" })
public class CreateDirectoryOverSFTP extends AppianSmartService {
	private static final Logger LOG = Logger
			.getLogger(CreateDirectoryOverSFTP.class);
	private final SmartServiceContext smartServiceCtx;
	private final ContentService contentService;
	private final SecureCredentialsStore scs;

	private final String STATUS_FAILED ="FAILED";
	private final String STATUS_SUCCESS = "SUCCESS";
	
	private String status;
	private String error;

	
	// Secure Credentials Store variables
	private SFTPConfig config = new SFTPConfig();
	private static final String PLUGIN_KEY = "com.appiancorp.a2.process.runtime.activities.sftp.CreateDirectoryOverSFTPwithSCSUsernameAndPassword";

	// Smart Service inputs
	private String remoteFilePath;
	private String newDirectoryName;

	public CreateDirectoryOverSFTP(SecureCredentialsStore scs,
			SmartServiceContext smartServiceCtx, ContentService cs) {
		super();
		this.scs = scs;
		this.smartServiceCtx = smartServiceCtx;
		this.contentService = cs;
	}

	private SmartServiceException createException(Throwable t, String key,
			Object... args) {
		return new SmartServiceException.Builder(getClass(), t).userMessage(
				key, args).build();
	}

	@Input(required = Required.ALWAYS)
	@Name("scsExternalSystemKey")
	public void setExternalSystemKey(String scsExternalSystemKey) {
		config.setScsExternalSystemKey(scsExternalSystemKey);
	}

	@Input(required = Required.ALWAYS, defaultValue = "false")
	@Name("usePerUserCredentials")
	public void setUsePerUserCredentials(Boolean usePerUserCredentials) {
		config.setUsePerUserCredentials(usePerUserCredentials);
	}

	@Input(required = Required.OPTIONAL)
	@Name("privateKey")
	@DocumentDataType
	public void setPrivateKeyFile(Long privateKeyDocId) {
		config.setPrivateKeyFileId(privateKeyDocId);
	}

	@Input(required = Required.OPTIONAL)
	@Name("knownHosts")
	@DocumentDataType
	public void setKnownHostsFile(Long knownHostsDocId) {
		config.setKnownHostsFileId(knownHostsDocId);
	}

	@Name("serverHostName")
	@Input(required = Required.ALWAYS)
	public void setServerHostName(String hostName) {
		config.setServerHostname(hostName);
	}

	@Name("serverPort")
	@Input(required = Required.ALWAYS, defaultValue = "22")
	public void setServerPort(Long portNumber) {
		config.setServerPort(portNumber);
	}

	@Name("remoteFilePath")
	@Input(required = Required.OPTIONAL)
	public void setRemoteFilePath(String serverFilepath) {
		this.remoteFilePath = serverFilepath;
	}

	@Name("newDirectoryName")
	@Input(required = Required.ALWAYS, defaultValue = "NewFolder")
	public void setNewDirectoryName(String newDirectoryName) {
		this.newDirectoryName = newDirectoryName;
	}

	@Name("Status")
	public String getStatus() {
		return status;
	}
	
	@Name("Error")
	public String getError() {
		return error;
	}
	
	@Override
	public void run() throws SmartServiceException {

		try {
			// Establish a new SFTP connection using process-specific parameters
			config.connect(contentService, scs);
		} catch (Exception e) {
			this.error=e.getMessage();
			this.status=STATUS_FAILED;
			LOG.error("Error connecting to remote SFTP server", e);
			throw createException(e, "error.sftp_conn", e.getMessage());
		}
		if (remoteFilePath != null && !remoteFilePath.isEmpty()) {
			Path path = Paths.get(remoteFilePath);
			Iterator<Path> folders = path.iterator();
			while (folders.hasNext()) {
				try {
					config.changeDirectory(folders.next().toString());
				} catch (Exception e) {
					LOG.error("Error changing the directory", e);
					this.error=e.getMessage();
					this.status=STATUS_FAILED;
					throw createException(e, "error.change_directory",
							e.getMessage());
				}
			}
		}
		try {
			// Create a directory on the remote server
			config.getSftpChannel().mkdir(newDirectoryName);
			this.status=STATUS_SUCCESS;

		} catch (Exception e) {
			LOG.error("Error creating folder in the SFTP server", e);
			this.error=e.getMessage();
			this.status=STATUS_FAILED;
			throw createException(e, "error.create_folder", e.getMessage());
		} finally {
			config.disconnect();
		}
	}

	/**
	 * Gets the corresponding message from the bundle for a given return code
	 * 
	 * @param code
	 *            the code returned from the executeTask() method
	 * @param params
	 *            the parameter values to replace
	 * @return the plain-text message
	 */
	protected String getMessageForCode(String code, Object... params) {
		Locale locale = smartServiceCtx.getUserLocale() == null ? smartServiceCtx
				.getPrimaryLocale() : smartServiceCtx.getUserLocale();

		try {
			ResourceBundle bundle = ResourceBundle
					.getBundle(PLUGIN_KEY, locale);

			return MessageFormat.format(bundle.getString(code), params);
		} catch (Exception e) {
			LOG.error("Error getting message for code " + code
					+ " from bundle " + PLUGIN_KEY, e);
			return "";
		}
	}
}
