package com.appian.ps.plugins.sftpsmartservices;

import java.io.FileInputStream;
import java.security.KeyStore;
import java.util.Map;
import org.apache.log4j.Logger;

import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.TrustManagerFactory;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.net.ftp.FTP;
import org.apache.commons.net.ftp.FTPReply;
import org.apache.commons.net.ftp.FTPSClient;
import org.apache.commons.net.util.TrustManagerUtils;

import com.appiancorp.suiteapi.content.ContentConstants;
import com.appiancorp.suiteapi.content.ContentService;
import com.appiancorp.suiteapi.security.external.SecureCredentialsStore;


public class FTPSConfig {
	public static final Integer SCS_ERROR=0x01;
	public static final Integer CONNECTION_ERROR = 0x02; 
	public static final Integer AUTHENTICATION_ERROR = 0x03;
	public static final Integer CERTIFICATE_ERROR = 0x04;
	public static final Integer GENERAL_ERROR = 0x05; 
	public static final Integer FILE_NOT_FOUND = 0x06; 
	public static final Integer STORE_FILE = 0x07;
	public static final Integer NON_POSITIVE_REPLY = 0x08;
	public static final Integer SOCKET_TIMEOUT = 0x09;
	
	public static final String[] errorMessages= {"NoError", 
			"SCS.error",
			"connection.error", 
			"authentication.error",
			"cert.error",
			"general.error",
			"file.error",
			"storefile.error",
			"nonpositive.error",
			"sockettimeout.error"
	};
	
	
	private static final Logger LOG=Logger.getLogger(FTPSConfig.class);
	private static final String KEY_USERNAME = "username";
	private static final String KEY_PASSWORD = "password";
	private static final String KEY_PASSPHRASE = "passphrase";
	private static final String KEY_NATPASSIVE = "natpassive"; 
	
	private String serverHostname;
	private Long serverPort;
	private String username;
	private String password;
	private String passphrase;
	private String scsExternalSystemKey;
	private boolean usePerUserCredentials = false;
	private Long privateKeyDocId;
	private boolean natpassive = false; 


	public String getServerHostname() {
		return serverHostname;
	}

	public Long getServerPort() {
		return serverPort;
	}

	public String getUsername() {
		return username;
	}

	public String getPassword() {
		return password;
	}
	
	public String getPassphrase() {
		return passphrase;
	}

	public String getScsExternalSystemKey() {
		return scsExternalSystemKey;
	}

	public boolean getUsePerUserCredentials() {
		return usePerUserCredentials;
	}

	public Long getPrivateKeyFileId() {
		return privateKeyDocId;
	}
	
	public boolean getNatpassive() {
		return natpassive; 
	}


	public void setServerHostname(String serverHostname) {
		this.serverHostname = serverHostname;
	}

	public void setServerPort(Long serverPort) {
		this.serverPort = serverPort;
	}

	public void setUsername(String val) {
		this.username = val;
	}

	public void setPassword(String val) {
		this.password = val;
	}

	public void setPassphrase(String val) {
		this.passphrase = val;
	}

	public void setScsExternalSystemKey(String val) {
		this.scsExternalSystemKey = val;
	}

	public void setUsePerUserCredentials(Boolean val) {
		this.usePerUserCredentials = val;
	}

	public void setPrivateKeyFileId(Long privateKeyDocId) {
		this.privateKeyDocId = privateKeyDocId;
	}
	
	public void setNatpassive(boolean natpassive) {
		this.natpassive = natpassive; 
	}

	public void getAuthenticationInformation(ContentService cs, SecureCredentialsStore scs) throws Exception {

		if (StringUtils.isNotEmpty(scsExternalSystemKey)) {
			try {
				Map<String, String> credentials = usePerUserCredentials
						? scs.getUserSecuredValues(scsExternalSystemKey)
						: scs.getSystemSecuredValues(scsExternalSystemKey);

				if (!credentials.containsKey(KEY_USERNAME)) {
					throw new Exception(String.format("Required field %s does not exist in SCS (%s)", KEY_USERNAME, scsExternalSystemKey));
				}
				

				if(!credentials.containsKey(KEY_PASSWORD)) { 
					throw new Exception(String.format("Required field %s does not exist in SCS (%s)",KEY_PASSWORD, scsExternalSystemKey));					
				}

				setUsername(credentials.get(KEY_USERNAME));
				setPassword(credentials.get(KEY_PASSWORD));
				setPassphrase(credentials.get(KEY_PASSPHRASE));
				setNatpassive("true".equalsIgnoreCase(credentials.get(KEY_NATPASSIVE)));
			} catch (Exception e) {
				throw new Exception(String.format("Unable to retrieve credentials from SCS: %s", e.getMessage()));
			}
		}
	}
	
	public FTPSClient setupConnection(ContentService cs) throws Exception {
		 LOG.debug("Configuration object created");
	      LOG.debug("Attempting to make connection to: "+getServerHostname());
	      
	      FTPSClient ftps = new FTPSClient("SSL");
	      // Establish a new SFTP connection using process-specific parameters
	      LOG.debug(getPrivateKeyFileId());
	      if (getPrivateKeyFileId() != null && getPassphrase()!=null) {
	    	    LOG.debug("Retrieveing Key Information");
				String privateKeyFilePath = cs.getInternalFilename(cs.getVersionId(getPrivateKeyFileId(), ContentConstants.VERSION_CURRENT));
				KeyStore ks = KeyStore.getInstance("JKS");
				try (FileInputStream fis = new FileInputStream(privateKeyFilePath)) { 
					ks.load(fis,getPassphrase().toCharArray()); 
				}
			    KeyManagerFactory kmf = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
			    kmf.init(ks,getPassphrase().toCharArray()); 
			    
			    TrustManagerFactory trustManagerFactory = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
			    trustManagerFactory.init(ks);
			    
			    ftps.setKeyManager(kmf.getKeyManagers()[0]);
			    ftps.setTrustManager(trustManagerFactory.getTrustManagers()[0]); 
			    ftps.setAuthValue("SSL");
			    
			    LOG.debug("Finished setting Key Information"); 
			
			} else { 
				ftps.setTrustManager(TrustManagerUtils.getAcceptAllTrustManager());
				ftps.setAuthValue("SSL");
			}
	       

	      ftps.setStrictReplyParsing(false);
	      LOG.debug("Calling remote server");
	      
	      if (natpassive) {
	    	  LOG.debug("Enabling Nat Passive Workaround Strategy");
	       	  ftps.setPassiveNatWorkaroundStrategy(new NatServerResolverImpl(this));
	      }
	      
	      return ftps; 
		
		
	}
	
	public void connect(FTPSClient ftps) throws Exception {
		      int reply; 
		      LOG.debug("Establishing Connection");
		      ftps.connect(getServerHostname(),getServerPort().intValue()); 
		      LOG.debug("Connection established"); 
		      LOG.debug("Getting Reply code");
		      reply = ftps.getReplyCode(); 
		      if (!FTPReply.isPositiveCompletion(reply)) {
		    	  ftps.disconnect();
		    	  throw new Exception("Error connection to remote FTPS Server with reply code: "+reply);
		      }
		      LOG.debug("Connection Successful");
		    LOG.debug("Connection established to "+getServerHostname()+" Preparing to attempt authentication"); 
		    
		    	ftps.setBufferSize(1000);
		    
		    	
		    	if (!ftps.login(getUsername(), getPassword())) {
		    		reply=ftps.getReplyCode(); 
		    		ftps.logout();
		    		throw new Exception("Unable to authenticate specified user, reply code: "+reply);  
		    	}
		    	
		    	// Set to binary transfer 
		    	ftps.setFileType(FTP.BINARY_FILE_TYPE);
		    	ftps.execPBSZ(0);
		    	ftps.execPROT("P");
		    	// Set to Passive Mode as behind firewall 
		    	ftps.enterLocalPassiveMode();
		    	
	}
}