package com.appian.ps.plugins.sftpsmartservices;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.MessageFormat;
import java.util.Locale;
import java.util.Map;
import java.util.ResourceBundle;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.appiancorp.suiteapi.common.Name;
import com.appiancorp.suiteapi.content.ContentConstants;
import com.appiancorp.suiteapi.content.ContentService;
import com.appiancorp.suiteapi.knowledge.Document;
import com.appiancorp.suiteapi.knowledge.DocumentDataType;
import com.appiancorp.suiteapi.knowledge.FolderDataType;
import com.appiancorp.suiteapi.process.exceptions.SmartServiceException;
import com.appiancorp.suiteapi.process.framework.AppianSmartService;
import com.appiancorp.suiteapi.process.framework.Input;
import com.appiancorp.suiteapi.process.framework.Order;
import com.appiancorp.suiteapi.process.framework.Required;
import com.appiancorp.suiteapi.process.framework.SmartServiceContext;
import com.appiancorp.suiteapi.process.palette.ConnectivityServices;
import com.appiancorp.suiteapi.security.external.SecureCredentialsStore;

@ConnectivityServices
@Order ({"scsExternalSystemKey", "usePerUserCredentials", "ServerHost", "Port", "RemoteFileUrl", "FolderToSave", "EnterLocalActiveMode", "BinaryFileType"})
public class ReceiveFileOverFTP extends AppianSmartService {

	private static final Logger LOG = Logger.getLogger(ReceiveFileOverFTP.class);
	private final SmartServiceContext smartServiceCtx;
	private final SecureCredentialsStore scs;
	private ContentService cs;
	
	// Smart Service inputs
	private String scsExternalSystemKey;
	private boolean usePerUserCredentials;
	private String serverHost;
	private Long port;
	private String remoteFileUrl;
	private Long folderToSave;
	private Boolean enterLocalActiveMode;
	private Boolean binaryFileType;
	
	// Smart Service outputs
	private Long doc;
	private String status;
	private String error;
	
	private Document d;
	
	private String STATUS_SUCCESS = "SUCCESS";
	private String STATUS_FAILED = "FAILED";
	private String STATUS_FILE_NOT_FOUND = "FILE_NOT_FOUND";
	// Secure Credentials Store variables
	private SFTPConfig config = new SFTPConfig();
	private static final String KEY_USERNAME = "username";
	private static final String KEY_PASSWORD = "password";
	private static final String NO_SCS_FIELD = "missing.scs.field";
	private static final String PLUGIN_KEY = "com.appiancorp.a2.process.runtime.activities.sftp.ReceiveFileOverFTP";	
	private String result;
	
	@Override
	public void run() throws SmartServiceException {
		// Use Secure Credential Store
	    scsExternalSystemKey = config.getScsExternalSystemKey();	    
	    usePerUserCredentials = config.getUsePerUserCredentials();
		if (StringUtils.isNotEmpty(scsExternalSystemKey)) {
			try {
				Map<String, String> credentials = usePerUserCredentials ? scs.getUserSecuredValues(scsExternalSystemKey) : scs.getSystemSecuredValues(scsExternalSystemKey);
				if (!credentials.containsKey(KEY_USERNAME)) {
					result = getMessageForCode(NO_SCS_FIELD, KEY_USERNAME, scsExternalSystemKey);
					throw new Exception(result);
				} else if (!credentials.containsKey(KEY_PASSWORD)) {
					result = getMessageForCode(NO_SCS_FIELD, KEY_PASSWORD, scsExternalSystemKey);
					throw new Exception(result);
				} 
				config.setUsername(credentials.get(KEY_USERNAME));
				config.setPassword(credentials.get(KEY_PASSWORD));
			} catch (Exception e) {
				LOG.error("Unable to retrieve credentials from SCS", e);
				this.status = STATUS_FAILED;
				this.error = e.getMessage();
				return;
			}
		}

		try {
			ftpTransfer(serverHost, config.getUsername(), config.getPassword(), remoteFileUrl, binaryFileType, enterLocalActiveMode);
			this.status=STATUS_SUCCESS;
		} catch (Exception e){
			LOG.error("Error transferring file", e);
			this.status = STATUS_FAILED;
			this.error = e.getMessage();
		}

	}

	public ReceiveFileOverFTP(SmartServiceContext smartServiceCtx, ContentService cs, SecureCredentialsStore scs) {
		super();
		this.scs = scs;
		this.smartServiceCtx = smartServiceCtx;
		this.cs = cs;
	}


	private boolean ftpTransfer(String server, String username, String password, String remoteFile, Boolean binaryTransferMode, Boolean enterLocalActiveMode ) throws Exception {

		try
		{
			
			remoteFileUrl = remoteFileUrl.replace('\\', '/');
			
		    String remoteFileName = remoteFileUrl.substring(remoteFileUrl.lastIndexOf('/') + 1);
		    
		    String remoteExtension = "";
		    if (remoteFileName.lastIndexOf('.') != -1) {
		      int dotIndex = remoteFileName.lastIndexOf('.');
		      remoteExtension = remoteFileName.substring(dotIndex + 1);
		      remoteFileName = remoteFileName.substring(0,dotIndex);
		    }
		    
		    LOG.debug("Remote File URL : " + remoteFileUrl);
		    LOG.debug("Remote File Name : " + remoteFileName);
		    LOG.debug("Remote File Extension : " + remoteExtension);
		    
		    String documentDescription = "Transferred via FTP";
		    
		    d = new Document(folderToSave, remoteFileName, remoteExtension);
	        d.setDescription(documentDescription);
	        d.setState(ContentConstants.STATE_PUBLISHED);
	        
	        LOG.debug("Appian Document " + d.getDisplayName());
	        LOG.debug("Appian Document " + d.getInternalFilename());
	        
	        // Attempt to create the document in Appian using the Content Service
	        Long documentId = cs.create(d, ContentConstants.UNIQUE_NONE);
	        
	        String internalFilePath = cs.getInternalFilename(documentId).replace('\\', '/');
	        String internalFileName;
	        
	        LOG.debug("Internal File Path - Friendly URL " + internalFilePath);
	        
	        internalFileName = internalFilePath.substring(internalFilePath.lastIndexOf('/') + 1);
	        internalFilePath = internalFilePath.substring(0, internalFilePath.lastIndexOf('/') + 1);
	        
	        LOG.debug("Internal File Name " + internalFileName);
	        LOG.debug("Internal File Path " + internalFilePath);
	        
	        try {
	          File internalFile = new File(internalFilePath);
	          if (!internalFile.exists())
	            internalFile.mkdirs();
	        } catch (Exception e) {
	        	LOG.error("Error creating folder path", e);
	        	this.error = e.getMessage();
	        	this.status= STATUS_FAILED;
	          throw e;
	        }
	     

	        FTPConfig.ftpTransferRecieveFile(server,port, username, password, remoteFileUrl, internalFilePath, internalFileName, binaryTransferMode, enterLocalActiveMode, cs, documentId);

	       
		

			// check for empty or file not found
			try {
				File f = new File(internalFilePath + internalFileName);
				LOG.debug("File size : " + f.length());
				LOG.debug("Document size in k : " + d.getSizeInKB());
				LOG.debug("Document : " + d);
				LOG.debug("Document Id : " + documentId);
				if (f.length() <= 0) {
					this.status = STATUS_FILE_NOT_FOUND;
					cs.delete(documentId, false);
					LOG.debug("Temp Appian document deleted "
							+ d.getExternalFilename());
				} else {
					this.status = STATUS_SUCCESS;
					this.doc = documentId;
				}

			} catch (Exception e) {
				LOG.error("Error deleting temporary document", e);
				this.error = e.getMessage();
				this.status = STATUS_FAILED;
				return false;
			}

		} catch (Exception ex) {
			LOG.error("Error while receiving file", ex);
			this.error = ex.getMessage();
			this.status = STATUS_FAILED;
			return false;
		}

		return true;
	}

	 /**
	   * Gets the corresponding message from the bundle for a given return code
	   * 
	   * @param code
	   *          the code returned from the executeTask() method
	   * @param params
	   *          the parameter values to replace
	   * @return the plain-text message
	   */
	 protected String getMessageForCode(String code, Object... params) {
	   Locale locale = smartServiceCtx.getUserLocale() == null ? smartServiceCtx.getPrimaryLocale()
	     : smartServiceCtx.getUserLocale();
	 
	   try {
	     ResourceBundle bundle = ResourceBundle.getBundle(PLUGIN_KEY, locale);

	     return MessageFormat.format(bundle.getString(code), params);
	   } catch (Exception e) {
	     LOG.error("Error getting message for code " + code + " from bundle " + PLUGIN_KEY, e);
	     return "";
	   }
	 }	
	 
	@Input(required = Required.ALWAYS)
	@Name("ServerHost")
	public void setServerHost(String val) {
		this.serverHost = val;
	}

	@Input(required = Required.ALWAYS)
	@Name("Port")
	public void setPort(Long val) {
		this.port = val;
	}
	
	@Input(required = Required.ALWAYS)
	@Name("scsExternalSystemKey")
	public void setExternalSystemKey(String scsExternalSystemKey) {
	   config.setScsExternalSystemKey(scsExternalSystemKey);
	}

	@Input(required = Required.ALWAYS, defaultValue = "false")
	@Name("usePerUserCredentials")
	public void setUsePerUserCredentials(Boolean usePerUserCredentials) {
	   config.setUsePerUserCredentials(usePerUserCredentials);
	}

	@Input(required = Required.ALWAYS)
	@Name("RemoteFileUrl")
	public void setRemoteFileUrl(String val) {
		this.remoteFileUrl = val;
	}

	@Input(required = Required.ALWAYS)
	@Name("FolderToSave")
	@FolderDataType
	public void setFolderToSave(Long val) {
		this.folderToSave = val;
	}

	@Input(required = Required.OPTIONAL, defaultValue = "false")
	@Name("EnterLocalActiveMode")
	public void setEnterLocalActiveMode(Boolean val) {
		this.enterLocalActiveMode = val;
	}

	@Input(required = Required.OPTIONAL, defaultValue = "true")
	@Name("BinaryFileType")
	public void setBinaryFileType(Boolean val) {
		this.binaryFileType = val;
	}

	@Name("doc")
	@DocumentDataType
	public Long getDoc() {
		return doc;
	}

	@Name("Status")
	public String getStatus() {
		return status;
	}

	@Name("Error")
	public String getError() {
		return error;
	}

}
