	// Authors: Phil Russell, Jed Fonner
package com.appian.ps.plugins.sftpsmartservices;

import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.MessageFormat;
import java.util.Locale;
import java.util.ResourceBundle;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.appiancorp.suiteapi.common.Name;
import com.appiancorp.suiteapi.content.Approval;
import com.appiancorp.suiteapi.content.ContentConstants;
import com.appiancorp.suiteapi.content.ContentService;
import com.appiancorp.suiteapi.knowledge.Document;
import com.appiancorp.suiteapi.knowledge.DocumentDataType;
import com.appiancorp.suiteapi.knowledge.FolderDataType;
import com.appiancorp.suiteapi.process.exceptions.SmartServiceException;
import com.appiancorp.suiteapi.process.framework.AppianSmartService;
import com.appiancorp.suiteapi.process.framework.Input;
import com.appiancorp.suiteapi.process.framework.MessageContainer;
import com.appiancorp.suiteapi.process.framework.Order;
import com.appiancorp.suiteapi.process.framework.Required;
import com.appiancorp.suiteapi.process.framework.SmartServiceContext;
import com.appiancorp.suiteapi.process.palette.ConnectivityServices;
import com.appiancorp.suiteapi.security.external.SecureCredentialsStore;

@ConnectivityServices
@Order({"scsExternalSystemKey", "usePerUserCredentials", "privateKey", "knownHosts", "serverHostName", "serverPort", "remoteFilePath", "destinationFolder", "destinationDocument", "documentName", "documentDescription"})
public class ReceiveFileOverSFTP extends AppianSmartService {

  private static final Logger LOG = Logger.getLogger(ReceiveFileOverSFTP.class);
  private final SmartServiceContext smartServiceCtx;
  private final ContentService contentService;
  private final SecureCredentialsStore scs;
  
  private static String INPUT_DESTINATION_FOLDER = "destinationFolder";
  private static String INPUT_DESTINATION_DOCUMENT = "destinationDocument";
  
  // Secure Credentials Store variables
  private SFTPConfig config = new SFTPConfig();
  private static final String PLUGIN_KEY = "com.appiancorp.a2.process.runtime.activities.sftp.ReceiveFileOverSFTPwithSCSUsernameAndPasswordV2";	

  // Smart Service inputs
  private String remoteFilePath;
  private Long destinationFolder;
  private Long destinationDocument;
  private String documentName;
  private String documentDescription;
  private Boolean deleteAfterReceive;

  private final String STATUS_FAILED ="FAILED";
  private final String STATUS_SUCCESS = "SUCCESS";

  // Smart Service outputs
  private Long newDocument;
  private String error;
  private String status;

  public ReceiveFileOverSFTP(SecureCredentialsStore scs, SmartServiceContext smartServiceCtx, ContentService cs) {
    super();
    this.scs = scs;
    this.smartServiceCtx = smartServiceCtx;
    this.contentService = cs;
  }

  @Override
  public void run() throws SmartServiceException {
    try {
      config.connect(contentService, scs);
    } catch (Exception e) {
      LOG.error("Error connecting to SFTP server", e);
      this.error=e.getMessage();
      this.status=STATUS_FAILED;
      throw createException(e, "error.sftp_conn", e.getMessage());
    }

    try {
      runSmartService();
      this.status=STATUS_SUCCESS;
    }
    catch (Exception e) {
        LOG.error("Error connecting to SFTP server", e);
        this.error=e.getMessage();
        this.status=STATUS_FAILED;
     
      }
    finally {
      config.disconnect();
    }
  }
  
  public void runSmartService() throws SmartServiceException {
    String internalFileName = null;
    String internalFilePath = null;
    Long documentId = null;
    Document d = null;

    // Convert user-provided pathnames for file system compatibility
    remoteFilePath = remoteFilePath.replace('\\', '/');
    String remoteFileName = remoteFilePath.substring(remoteFilePath.lastIndexOf('/') + 1);

    // Set all required parameters for the Appian Document
    // If the documentname_param field is not set, set it to the remote_filename variable
    // Also, check to make sure that, if the user specifies a different document name, that
    // they included a file extension. If not, duplicate file extension from remote_filename.
    if (documentName == null || documentName.length() == 0) {
      documentName = remoteFileName;
    }

    // get the remote extension so we can append it to the document name if it was omitted
    String remoteExtension = "";
    if (remoteFileName.lastIndexOf('.') != -1) {
      int dotIndex = remoteFileName.lastIndexOf('.');
      remoteExtension = remoteFileName.substring(dotIndex + 1);
    }
    // separate the document name and extension for the API call
    String documentExtension = "";
    if (documentName.lastIndexOf('.') != -1) {
      int dotIndex = documentName.lastIndexOf('.');
      documentExtension = documentName.substring(dotIndex + 1);
      documentName = documentName.substring(0, dotIndex);
    }
    // if no document extension was specified but the remote file has an extension, use remote extension
    if (StringUtils.isBlank(documentExtension) && StringUtils.isNotBlank(remoteExtension)) {
      documentExtension = remoteExtension;
    } else if (!documentExtension.equalsIgnoreCase(remoteExtension)) {
      LOG.warn("Attempting to transfer file with extension '" + remoteExtension +
        "' to Document with extension '" + documentExtension + "'");
    }

    // Set the document description if the user does not provide one
    if (StringUtils.isBlank(documentDescription)) {
      documentDescription = "Transferred via SFTP";
    }
    try {
      if (null == destinationDocument) {
        // Create the new Appian document. Then set the description and status of the document.
        d = new Document(destinationFolder, documentName, documentExtension);
        d.setDescription(documentDescription);
        d.setState(ContentConstants.STATE_PUBLISHED);
        // Attempt to create the document in Appian using the Content Service
        documentId = contentService.create(d, ContentConstants.UNIQUE_NONE);
      } else {
        Document doc = (Document) contentService.getVersion(destinationDocument,
          ContentConstants.VERSION_CURRENT);
        // prepare the document, indicating the service to assign a new file path
        // to the new version of the document
        doc.setFileSystemId(ContentConstants.ALLOCATE_FSID);
        // create a new version
        Approval a = contentService.createVersion(doc, ContentConstants.UNIQUE_NONE);
        // get the file path of the new version
        documentId = a.getId()[0];
      }

      // Get local filepath on Appian server where physical file was created
      internalFilePath = contentService.getInternalFilename(documentId).replace('\\', '/');
    } catch (Exception e) {
      LOG.error("Error creating Appian document", e);
      throw createException(e, "error.file_create", e.getMessage());
    }

    // Set filename and full path respectively
    internalFileName = internalFilePath.substring(internalFilePath.lastIndexOf('/') + 1);
    internalFilePath = internalFilePath.substring(0, internalFilePath.lastIndexOf('/') + 1);

    try {
      // Check to see if local directory exists. If it doesn't try to create it.
      // Then, change to local directory where ACC document's internal filename is pointing
      File internalFile = new File(internalFilePath);
      if (!internalFile.exists())
        internalFile.mkdirs();
    } catch (Exception e) {
      LOG.error("Error creating Appian document", e);
      throw createException(e, "error.file_create", e.getMessage());
    }

    // Get remote file
    try (InputStream fileBodyIns = config.getSftpChannel().get(remoteFilePath);
    		
       // BufferedOutputStream out = new BufferedOutputStream(new FileOutputStream(new File(internalFilePath + internalFileName)))
    ) {
    	File fh = new File(internalFilePath + internalFileName);
		OutputStream out = new FileOutputStream(fh);
      // According to SFTP protocol, 32768 is the largest possible read:
      // https://tools.ietf.org/html/draft-ietf-secsh-filexfer-13#section-4
      // "The maximum size of a packet is in practice determined by the
      // client (the maximum size of read or write requests that it sends,
      // plus a few bytes of packet overhead).  All servers SHOULD support
      // packets of at least 34000 bytes (where the packet size refers to
      // the full length, including the header above).  This should allow
      // for reads and writes of at most 32768 bytes."
      //byte buf[] = new byte[32768];
    	//1024
    	byte buf[] = new byte[1024];
      int len;

      // Write remote file to local disk
      while ((len = fileBodyIns.read(buf)) > 0) {
        out.write(buf, 0, len);
      }
      
      out.close();

      // Update the Appian document metadata with the file size
      contentService.setSizeOfDocumentVersion(documentId);
    } catch (Exception e) {
      LOG.error("Error transferring file", e);
      throw createException(e, "error.file_transfer", e.getMessage());
    }

    if (deleteAfterReceive) {
      try {
        LOG.info("Removing file from remote server at " + remoteFilePath);
        config.getSftpChannel().rm(remoteFilePath);
      } catch (Exception e) {
        LOG.error("Error transferring file", e);
        throw createException(e, "error.file_transfer", e.getMessage());
      }
    }

    // Set the "newDocument" ARV to the documentID of the new document created
    newDocument = documentId;
  }

  private SmartServiceException createException(Throwable t, String key, Object... args) {
    return new SmartServiceException.Builder(getClass(), t).userMessage(key, args).build();
  }

  /**
   * Gets the corresponding message from the bundle for a given return code
   * 
   * @param code
   *          the code returned from the executeTask() method
   * @param params
   *          the parameter values to replace
   * @return the plain-text message
   */
  protected String getMessageForCode(String code, Object... params) {
    Locale locale = smartServiceCtx.getUserLocale() == null ? smartServiceCtx.getPrimaryLocale()
     : smartServiceCtx.getUserLocale();
 
    try {
      ResourceBundle bundle = ResourceBundle.getBundle(PLUGIN_KEY, locale);

      return MessageFormat.format(bundle.getString(code), params);
    } catch (Exception e) {
      LOG.error("Error getting message for code " + code + " from bundle " + PLUGIN_KEY, e);
      return "";
    }
  }  

  public void validate(MessageContainer messages) {
    if (null == destinationFolder && null == destinationDocument) {
      messages.addError(INPUT_DESTINATION_FOLDER, "error.missing_folder_or_doc");
    }
    if (null != destinationFolder && null != destinationDocument) {
      messages.addError(INPUT_DESTINATION_FOLDER, "error.only_specify_folder_or_doc");
    }
    if (null != destinationDocument && StringUtils.isNotBlank(documentName)) {
      try {
        Document d = (Document) contentService.getVersion(destinationDocument,
          ContentConstants.VERSION_CURRENT);
        String ext = d.getExtension();
        int dotIndex = documentName.lastIndexOf('.');
        String documentExtension = documentName.substring(dotIndex + 1);
        if (!ext.equalsIgnoreCase(documentExtension)) {
          messages.addError(INPUT_DESTINATION_DOCUMENT, "error.extension_mismatch");
        }
      } catch (Exception e) {
        messages.addError(INPUT_DESTINATION_DOCUMENT, "error.insufficient_permission");
      }
    }
  }
  
  @Input(required = Required.ALWAYS)
  @Name("scsExternalSystemKey")
  public void setExternalSystemKey(String scsExternalSystemKey) {
     config.setScsExternalSystemKey(scsExternalSystemKey);
  }

  @Input(required = Required.ALWAYS, defaultValue = "false")
  @Name("usePerUserCredentials")
  public void setUsePerUserCredentials(Boolean usePerUserCredentials) {
     config.setUsePerUserCredentials(usePerUserCredentials);
  }

  @Input(required = Required.OPTIONAL)
  @Name("privateKey")
  @DocumentDataType
  public void setPrivateKeyFile(Long privateKeyDocId) {
     config.setPrivateKeyFileId(privateKeyDocId);
  }

  @Input(required = Required.OPTIONAL)
  @Name("knownHosts")
  @DocumentDataType
  public void setKnownHostsFile(Long knownHostsDocId) {
     config.setKnownHostsFileId(knownHostsDocId);
  }

  @Input(required = Required.ALWAYS)
  @Name("serverHostName")
  public void setServerHostName(String val) {
    config.setServerHostname(val);
  }

  @Input(required = Required.ALWAYS, defaultValue = "22")
  @Name("serverPort")
  public void setServerPort(Long val) {
    config.setServerPort(val);
  } 
  @Input(required = Required.ALWAYS)
  @Name("remoteFilePath")
  public void setRemoteFilePath(String val) {
    this.remoteFilePath = val;
  }

  @Input(required = Required.OPTIONAL)
  @Name("destinationFolder")
  @FolderDataType
  public void setDestinationFolder(Long val) {
    this.destinationFolder = val;
  }

  @Input(required = Required.OPTIONAL)
  @Name("destinationDocument")
  @DocumentDataType
  public void setDestinationDocument(Long val) {
    this.destinationDocument = val;
  }

  @Input(required = Required.OPTIONAL)
  @Name("documentName")
  public void setDocumentName(String val) {
    this.documentName = val;
  }

  @Input(required = Required.OPTIONAL)
  @Name("documentDescription")
  public void setDocumentDescription(String val) {
    this.documentDescription = val;
  }

  @Input(required = Required.OPTIONAL, defaultValue = "false")
  @Name("deleteAfterReceive")
  public void setDeleteAfterReceive(Boolean val) {
    this.deleteAfterReceive = val;
  }

  @Name("newDocument")
  @DocumentDataType
  public Long getNewDocument() {
    return newDocument;
  }

	@Name("Error")
	public String getError() {
		return error;
	}
	
	@Name("Status")
	public String getStatus() {
		return status;
	}
}
