// Author: Phil Russell
// Date Modified: 2010-August-19
// Appian Version: 6.0.3+

package com.appian.ps.plugins.sftpsmartservices;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.text.MessageFormat;
import java.util.Locale;
import java.util.ResourceBundle;

import org.apache.commons.net.ftp.FTPConnectionClosedException;
import org.apache.commons.net.ftp.FTPReply;
import org.apache.commons.net.ftp.FTPSClient;
import org.apache.log4j.Logger;

import com.appiancorp.suiteapi.common.Name;
import com.appiancorp.suiteapi.content.ContentConstants;
import com.appiancorp.suiteapi.content.ContentService;
import com.appiancorp.suiteapi.content.exceptions.InvalidContentException;
import com.appiancorp.suiteapi.knowledge.DocumentDataType;
import com.appiancorp.suiteapi.process.exceptions.SmartServiceException;
import com.appiancorp.suiteapi.process.framework.AppianSmartService;
import com.appiancorp.suiteapi.process.framework.Input;
import com.appiancorp.suiteapi.process.framework.Order;
import com.appiancorp.suiteapi.process.framework.Required;
import com.appiancorp.suiteapi.process.framework.SmartServiceContext;
import com.appiancorp.suiteapi.process.palette.ConnectivityServices;
import com.appiancorp.suiteapi.security.external.SecureCredentialsStore;

@ConnectivityServices
@Order({"scsExternalSystemKey", "usePerUserCredentials", "privateKey", "serverHostName", "serverPort", "remoteFilePath", "sourceDocument"})
public class SendFileOverFTPS extends AppianSmartService {
  private static final Logger LOG = Logger.getLogger(SendFileOverFTPS.class);
  private final SmartServiceContext smartServiceCtx;
  private final ContentService contentService;
  private final SecureCredentialsStore scs;
  private FTPSClient ftps; 
  private String[] errorMessage;
  private Boolean isSuccessful = true; 
  private int errorCode; 
  private static final int TIME_OUT = 15*60*1000; /* Milliseconds */ 
  private static final int KEEP_ALIVE = 20; /* Seconds */ 

  // Secure Credentials Store variables
  private FTPSConfig config = new FTPSConfig();
  private static final String PLUGIN_KEY = "com.appiancorp.a2.process.runtime.activities.sftp.SendFileOverFTPS";	

  // Smart Service inputs
  private String remoteFilePath;
  private Long sourceDocument;

  public SendFileOverFTPS(SecureCredentialsStore scs, SmartServiceContext smartServiceCtx, ContentService cs) {
    super();
    this.scs = scs;
    this.smartServiceCtx = smartServiceCtx;
    this.contentService = cs;
  }

  @Input(required = Required.ALWAYS)
  @Name("scsExternalSystemKey")
  public void setExternalSystemKey(String scsExternalSystemKey) {
     config.setScsExternalSystemKey(scsExternalSystemKey);
  }
  
  @Input(required = Required.ALWAYS, defaultValue = "false")
  @Name("usePerUserCredentials")
  public void setUsePerUserCredentials(Boolean usePerUserCredentials) {
     config.setUsePerUserCredentials(usePerUserCredentials);
  }  

  @Input(required = Required.OPTIONAL)
  @Name("privateKey")
  @DocumentDataType
  public void setPrivateKeyFile(Long privateKeyDocId) {
     config.setPrivateKeyFileId(privateKeyDocId);
  }

  @Name("serverHostName")
  @Input(required = Required.ALWAYS)
  public void setServerHostName(String hostName) {
    config.setServerHostname(hostName);
  }

  @Name("serverPort")
  @Input(required = Required.ALWAYS, defaultValue = "21")
  public void setServerPort(Long portNumber) {
    config.setServerPort(portNumber);
  }

  @Name("remoteFilePath")
  @Input(required = Required.ALWAYS)
  public void setRemoteFilePath(String serverFilepath) {
    this.remoteFilePath = serverFilepath;
  }

  @Name("sourceDocument")
  @Input(required = Required.ALWAYS)
  @DocumentDataType
  public void setSourceDocument(Long documentId) {
    this.sourceDocument = documentId;
  }
  
  
  /* Outputs */ 
  
	public String[] getErrorMessage() {
		return errorMessage;
	}
	
	public Boolean getIsSuccessful() {
		return isSuccessful;
	}
	
	public int getErrorCode() {
		return errorCode;
	}

  @Override
  public void run() throws SmartServiceException {
    String internalFilePath; 
    
    try {
      // Obtain the current version Document object of the ACC document specified
      sourceDocument = contentService.getVersionId(sourceDocument, ContentConstants.VERSION_CURRENT);
      internalFilePath = contentService.getInternalFilename(sourceDocument).replace('\\', '/');
    } catch (Exception e) {
      this.setErrorMessageForCode(FTPSConfig.FILE_NOT_FOUND, e);
      LOG.error("Error obtaining document to send path", e);

      // The documentID specified does not exist in the Appian Collaboration Center
      return;
    }
    
    try { 
    	LOG.debug("Attempting to access the secure credential store for key: "+config.getScsExternalSystemKey());
        config.getAuthenticationInformation(contentService, scs);
    } catch (Exception e) { 
    	setErrorMessageForCode(FTPSConfig.SCS_ERROR,e); 
    	LOG.error(e); 
    	return; 
    }
    
    try { 
    	LOG.debug("Setting up the connection parameters with the keystore file"); 
    	ftps = config.setupConnection(contentService); 
        
    } catch (Exception e) {
    	this.setErrorMessageForCode(FTPSConfig.CERTIFICATE_ERROR, e);
    	LOG.error("Error in setting up connection parameters",e);
    	return; 
    }
    
    try {
    	config.connect(ftps);
		ftps.setSoTimeout(TIME_OUT);
		ftps.setControlKeepAliveTimeout(KEEP_ALIVE);
    } catch (Exception e) {
    	this.setErrorMessageForCode(FTPSConfig.CONNECTION_ERROR, e,config.getServerHostname());
    	LOG.error("Connection error", e);
    	return; 
    }
    
    LOG.debug("Opening Filestream");
    try (InputStream fileBodyIns = new FileInputStream(internalFilePath)){
    	LOG.debug("Transfering file"); 
    	String filename = contentService.getExternalFilename(sourceDocument); 
    	ftps.storeFile(remoteFilePath.concat("/").concat(filename), fileBodyIns);
    	int reply = ftps.getReplyCode(); 
    	if (reply == 450) {
    		this.setErrorMessageForCode(FTPSConfig.STORE_FILE,null,filename);
    		LOG.error("File action was not executed on the remote server - check destination directory exists and user has permissions");
    	} else
        if (!FTPReply.isPositiveCompletion(reply)) {
	    	  this.setErrorMessageForCode(FTPSConfig.NON_POSITIVE_REPLY, null, reply);
	    	  LOG.error("Non positive reply code returned "+reply);
	    }
    	LOG.debug("File Transfer Complete"); 
    	ftps.logout();
    } catch (FTPConnectionClosedException e)  {
        this.setErrorMessageForCode(FTPSConfig.GENERAL_ERROR, e);
    	LOG.error("Error FTP Connection closed",e);
    } catch (IOException e) {
    	this.setErrorMessageForCode(FTPSConfig.GENERAL_ERROR, e);
    	isSuccessful = false;
    	LOG.error("Error Transmitting file: ",e);
    } catch (InvalidContentException e) {
    	this.setErrorMessageForCode(FTPSConfig.GENERAL_ERROR, e);
    	LOG.error("Error Retrieving Source File" );
    } catch (Exception e) {
    	this.setErrorMessageForCode(FTPSConfig.GENERAL_ERROR, e);
    	LOG.error(e);
	} finally {
    
    	if (ftps.isConnected())
    	{
    		try {
    			ftps.disconnect();
    			
    		} catch (IOException f) {
    			// do nothing 
    		}
    		
    	}
    }
  
  }



/**
   * Gets the corresponding message from the bundle for a given return code
   * 
   * @param code
   *          the code returned from the executeTask() method
   * @param params
   *          the parameter values to replace
   * @return the plain-text message
   */
  protected void setErrorMessageForCode(Integer code, Exception ex, Object... params) {
	   errorCode = code.intValue(); 
	   isSuccessful = false; 
	   Locale locale = smartServiceCtx.getUserLocale() == null ? smartServiceCtx.getPrimaryLocale()
	     : smartServiceCtx.getUserLocale();
	 
	   try {
	     ResourceBundle bundle = ResourceBundle.getBundle(PLUGIN_KEY, locale);

	     if (ex != null) { 
	    	 errorMessage = new String[]{
	     
	     MessageFormat.format(bundle.getString(FTPSConfig.errorMessages[code.intValue()]), params)
	     ,ex.getMessage() }; 
	     } else { 
	    	 errorMessage = new String[] {MessageFormat.format(bundle.getString(FTPSConfig.errorMessages[code.intValue()]), params)};
	    	
	     }
	    	 
	   } catch (Exception e) {
		 errorMessage = new String[] {"Check Log for error"}; 
	     LOG.error("Error getting message for code " + code + " from bundle " + PLUGIN_KEY, e);
	     return;
	   }
  }    
}