// Author: Henry Chandra
// Date Modified: 2014-11-20
// Appian Version: 7.5+

package com.appian.ps.plugins.sftpsmartservices;

import java.text.MessageFormat;
import java.util.Locale;
import java.util.ResourceBundle;

import org.apache.log4j.Logger;

import com.appiancorp.suiteapi.common.Name;
import com.appiancorp.suiteapi.content.ContentService;
import com.appiancorp.suiteapi.knowledge.DocumentDataType;
import com.appiancorp.suiteapi.process.exceptions.SmartServiceException;
import com.appiancorp.suiteapi.process.framework.AppianSmartService;
import com.appiancorp.suiteapi.process.framework.Input;
import com.appiancorp.suiteapi.process.framework.Order;
import com.appiancorp.suiteapi.process.framework.Required;
import com.appiancorp.suiteapi.process.framework.SmartServiceContext;
import com.appiancorp.suiteapi.process.palette.ConnectivityServices;
import com.appiancorp.suiteapi.security.external.SecureCredentialsStore;

@ConnectivityServices
@Order({"scsExternalSystemKey", "usePerUserCredentials", "privateKey", "knownHosts", "serverHostName", "serverPort", "remoteFilePath"})
public class RemoveFileOverSFTP extends AppianSmartService {

	private static final Logger LOG = Logger.getLogger(RemoveFileOverSFTP.class);
	private final SmartServiceContext smartServiceCtx;
	private final SecureCredentialsStore scs;
	private final ContentService cs;

	// Secure Credentials Store variables
	private SFTPConfig config = new SFTPConfig();
	private static final String PLUGIN_KEY = "com.appiancorp.a2.process.runtime.activities.sftp.RemoveFileOverSFTPwithSCSUsernameAndPasswordV2";	

	// Smart Service inputs
	private String remoteFilePath;

	public RemoveFileOverSFTP(SecureCredentialsStore scs, SmartServiceContext smartServiceCtx, ContentService cs) {
		super();
		this.scs = scs;
		this.smartServiceCtx = smartServiceCtx;
		this.cs = cs;
	}

	@Override
	public void run() throws SmartServiceException {
		try {
			// Establish a new SFTP connection using process-specific parameters
			config.connect(cs, scs);
		} catch (Exception e) {
			LOG.error("Error connecting to SFTP server", e);
			throw createException(e, "error.sftp_conn", e.getMessage());
		}

		//Remove file over SFTP
		try {
			config.getSftpChannel().rm(remoteFilePath);
		} catch (Exception e) {
			LOG.error("Error removing file", e);
			throw createException(e, "error.file_remove", e.getMessage());
		}

		config.disconnect();
	}

	/**
	 * Gets the corresponding message from the bundle for a given return code
	 * 
	 * @param code
	 *          the code returned from the executeTask() method
	 * @param params
	 *          the parameter values to replace
	 * @return the plain-text message
	 */
	protected String getMessageForCode(String code, Object... params) {
		Locale locale = smartServiceCtx.getUserLocale() == null ? smartServiceCtx.getPrimaryLocale()
				: smartServiceCtx.getUserLocale();

		try {
			ResourceBundle bundle = ResourceBundle.getBundle(PLUGIN_KEY, locale);

			return MessageFormat.format(bundle.getString(code), params);
		} catch (Exception e) {
			LOG.error("Error getting message for code " + code + " from bundle " + PLUGIN_KEY, e);
			return "";
		}
	}
	
	private SmartServiceException createException(Throwable t, String key, Object... args) {
		return new SmartServiceException.Builder(getClass(), t).userMessage(key, args).build();
	}
	
	@Name("serverHostName")
	@Input(required = Required.ALWAYS)
	public void setServerHostName(String hostName) {
		config.setServerHostname(hostName);
	}

	@Name("serverPort")
	@Input(required = Required.ALWAYS, defaultValue = "22")
	public void setServerPort(Long portNumber) {
		config.setServerPort(portNumber);
	}

	@Input(required = Required.ALWAYS)
	@Name("scsExternalSystemKey")
	public void setExternalSystemKey(String scsExternalSystemKey) {
		config.setScsExternalSystemKey(scsExternalSystemKey);
	}

	@Input(required = Required.ALWAYS, defaultValue = "false")
	@Name("usePerUserCredentials")
	public void setUsePerUserCredentials(Boolean usePerUserCredentials) {
		config.setUsePerUserCredentials(usePerUserCredentials);
	}  

	@Input(required = Required.OPTIONAL)
	@Name("privateKey")
	@DocumentDataType
	public void setPrivateKeyFile(Long privateKeyDocId) {
		config.setPrivateKeyFileId(privateKeyDocId);
	}

	@Input(required = Required.OPTIONAL)
	@Name("knownHosts")
	@DocumentDataType
	public void setKnownHostsFile(Long knownHostsDocId) {
		config.setKnownHostsFileId(knownHostsDocId);
	}

	@Name("remoteFilePath")
	@Input(required = Required.ALWAYS)
	public void setRemoteFilePath(String serverFilepath) {
		this.remoteFilePath = serverFilepath;
	}
}
