package com.appian.ps.plugins.sftpsmartservices;

import java.io.FileInputStream;
import java.io.InputStream;

import org.apache.log4j.Logger;

import com.appiancorp.suiteapi.common.Name;
import com.appiancorp.suiteapi.content.ContentConstants;
import com.appiancorp.suiteapi.content.ContentService;
import com.appiancorp.suiteapi.knowledge.Document;
import com.appiancorp.suiteapi.knowledge.DocumentDataType;
import com.appiancorp.suiteapi.process.exceptions.SmartServiceException;
import com.appiancorp.suiteapi.process.framework.AppianSmartService;
import com.appiancorp.suiteapi.process.framework.Input;
import com.appiancorp.suiteapi.process.framework.Order;
import com.appiancorp.suiteapi.process.palette.ConnectivityServices;
import com.jcraft.jsch.ChannelSftp;
import com.jcraft.jsch.JSch;
import com.jcraft.jsch.Session;

@ConnectivityServices
@Order ({"serverHostName", "serverPort", "sftpUsername", "sftpPassword", "sftpServerFilePath", "sftpServerFileName", "documentToCopy"})
public class SendFileOverSFTPBasic extends AppianSmartService {
	private static final Logger LOG = Logger.getLogger(SendFileOverSFTPBasic.class);
	private final ContentService contentService;

	private Long documentToCopy;
	private String serverHostName;
	private Long serverPort;
	private String sftpUsername;
	private String sftpPassword;
	private String sftpServerFilePath;
	private String sftpServerFileName;

	@Override
	public void run() throws SmartServiceException {

		String local_filename = null;
	    String local_pathname = null;
	    String remote_filename = null;

	    Document d = null;

	    JSch jsch = null;
	    Session session = null;
	    ChannelSftp sftpChannel = null;

	    try {
	      // Obtain the current version Document object of the ACC document specified
	      d = contentService.download(documentToCopy, ContentConstants.VERSION_CURRENT, false)[0];
	    } catch (Exception e) {
	      LOG.error("Error getting current document", e);
	      // The documentID specified does not exist in the Appian Collaboration Center
	      throw createException(e, "error.doc_no_exist", e.getMessage());
	    }

	    // Get the local file and path information for the specified document
	    local_pathname = d.getInternalFilename().replace('\\','/');
	    local_filename = local_pathname.substring(local_pathname.lastIndexOf('/')+1);
	    local_pathname = local_pathname.substring(0, local_pathname.lastIndexOf('/')+1);

	    // Set the name of the remote file to transfer
	    remote_filename = d.getName() + "." + d.getExtension();
	    if(sftpServerFileName.length() > 0) {
	      remote_filename = sftpServerFileName;
	    }

	    // Ensure that the remote file path terminates with a forward slash
	    if(sftpServerFilePath.charAt(sftpServerFilePath.length()-1) != '/') {
	      sftpServerFilePath = sftpServerFilePath + '/';
	    }

    try {
      try {
        // Establish a new SFTP connection using process-specific parameters
        jsch = new JSch();
        session = jsch.getSession(sftpUsername, serverHostName);
        session.setPassword(sftpPassword);
        session.setPort(serverPort.intValue());
        session.setConfig("StrictHostKeyChecking", "no");
        session
            .setConfig(
                "cipher.s2c",
                "aes128-ctr,aes128-cbc,3des-ctr,3des-cbc,blowfish-cbc,aes192-cbc,aes256-cbc,aes256-ctr");
        session
            .setConfig(
                "cipher.c2s",
                "aes128-ctr,aes128-cbc,3des-ctr,3des-cbc,blowfish-cbc,aes192-cbc,aes256-cbc,aes256-ctr");
        session.connect();

        sftpChannel = (ChannelSftp) session.openChannel("sftp");
        sftpChannel.connect();
      } catch (Exception e) {
        LOG.error("Error connecting to SFTP server", e);
        throw createException(e, "error.sftp_conn", e.getMessage());
      }

      try {
        // Create a copy of the Appian document on the remote server
        InputStream fileBodyIns = new FileInputStream(local_pathname
            + local_filename);
        sftpChannel.put(fileBodyIns, sftpServerFilePath + remote_filename);
      } catch (Exception e) {
        LOG.error("Error transferring file", e);
        throw createException(e, "error.file_transfer", e.getMessage());
      }
    } finally {
      // Close SFTP channel and session
      if (sftpChannel != null) {
        sftpChannel.disconnect();
      }
      if (session != null) {
        session.disconnect();
      }
    }
	}

	public SendFileOverSFTPBasic(ContentService cs) {
		super();
		this.contentService = cs;
	}

	@Input(localId="var_local_id_0", required=com.appiancorp.suiteapi.process.framework.Required.ALWAYS)
	@Name("documentToCopy")
	@DocumentDataType
	public void setDocumentToCopy(Long val) {
		this.documentToCopy = val;
	}

	@Input(localId="var_local_id_1", required=com.appiancorp.suiteapi.process.framework.Required.ALWAYS)
	@Name("serverHostName")
	public void setServerHostName(String val) {
		this.serverHostName = val;
	}

	@Input(localId="var_local_id_2", required=com.appiancorp.suiteapi.process.framework.Required.ALWAYS, defaultValue="22")
	@Name("serverPort")
	public void setServerPort(Long val) {
		this.serverPort = val;
	}

	@Input(localId="var_local_id_3", required=com.appiancorp.suiteapi.process.framework.Required.ALWAYS)
	@Name("sftpUsername")
	public void setSftpUsername(String val) {
		this.sftpUsername = val;
	}

	@Input(localId="var_local_id_4", required=com.appiancorp.suiteapi.process.framework.Required.ALWAYS)
	@Name("sftpPassword")
	public void setSftpPassword(String val) {
		this.sftpPassword = val;
	}

	@Input(localId="var_local_id_5", required=com.appiancorp.suiteapi.process.framework.Required.ALWAYS)
	@Name("sftpServerFilePath")
	public void setSftpServerFilePath(String val) {
		this.sftpServerFilePath = val;
	}

	@Input(localId="var_local_id_6", required=com.appiancorp.suiteapi.process.framework.Required.OPTIONAL)
	@Name("sftpServerFileName")
	public void setSftpServerFileName(String val) {
		this.sftpServerFileName = val;
	}

	private SmartServiceException createException(Throwable t, String key, Object... args) {
		return new SmartServiceException.Builder(getClass(), t).userMessage(key, args).build();
	}
}
